#!/usr/bin/python
import sys
import unittest
from pyVmomi import Vim, Vmodl
from pyVim.connect import SmartConnect, Disconnect
from optparse import OptionParser
from pyVim import folder
from pyVim import host

"""
   Host Firewall System VIM API Test

   Usage:

   ../py.sh TestFirewallSystem.py -H <host name/ip> -u <user name> -p <password>

"""

def get_options():
    """
    Supports the command-line arguments listed below
    """

    parser = OptionParser()
    parser.add_option("-H", "--host",
                      default="localhost",
                      help="remote host to connect to")
    parser.add_option("-u", "--user",
                      default="root",
                      help="User name to use when connecting to hostd")
    parser.add_option("-p", "--password",
                      default="",
                      help="Password to use when connecting to hostd")
    (options, _) = parser.parse_args()
    return options


class TestFirewallSystem(unittest.TestCase):

   def setOptions(self, options):
      """
      Command line options
      """
      self.options = options

   def setUp(self):
      """
      Setting test suite
      """
      options = get_options()

      self.si = SmartConnect(host=self.options.host,
                             user=self.options.user,
                             pwd=self.options.password)
      configManager = host.GetHostSystem(self.si).configManager
      self.firewallSystem = configManager.firewallSystem

   def tearDown(self):
      """
      Reset test suite
      """
      Disconnect(self.si)

   def runTest(self):
      self.TestGetFirewallInfo()
      self.TestUpdateDefaultPolicy()
      self.TestUpdateRulesetStatus()
      self.TestUpdateRulesetAllowedHosts()

   def TestGetFirewallInfo(self):
      """
      Verify getfirewallinfo operation.
      """
      print("\nTest Firewall GetFirewallInfo API")

      self.assertTrue(self.firewallSystem != None,
                      "firewall system should not be None")

      firewallInfo = self.firewallSystem.firewallInfo
      self.assertTrue(firewallInfo != None,
                      "firewall Info should not be None")

      defaultPolicy = firewallInfo.defaultPolicy
      self.assertTrue(defaultPolicy != None,
                      "firewall default policy should not be None")
      self.assertTrue(defaultPolicy.incomingBlocked == defaultPolicy.outgoingBlocked,
                      "firewall default policy for incoming and outgoing should be same")

      rulesets = firewallInfo.ruleset
      self.assertTrue(rulesets != None,
                      "firewall rulesets should not be None")
      self.assertTrue(len(rulesets) > 0,
                      "firewall rulesets should not be empty")
      # check the required and enabled ruleset
      found = False
      for ruleset in rulesets:
         if ruleset.key == "vSphereClient":
            found = True
            self.assertTrue(ruleset.enabled and ruleset.required,
                           "vSphereClient ruleset enabled and required should not be false")
            break

      self.assertTrue(found,
                     "Not found vSphereClient ruleset")

      print("Test Firewall GetFirewallInfo: PASS")

   def TestUpdateDefaultPolicy(self):
      """
      General checks UpdateDefaultPolicy API
      """
      print("Test Firewall UpdateDefaultPolicy API")

      self.assertTrue(self.firewallSystem != None,
                      "firewall system should not be None")

      currentPolicy = self.firewallSystem.firewallInfo.defaultPolicy
      blocked = currentPolicy.incomingBlocked
      updatePolicy = Vim.Host.FirewallInfo.DefaultPolicy()
      newBlocked = (not blocked)
      updatePolicy.incomingBlocked = newBlocked
      updatePolicy.outgoingBlocked = newBlocked

      error = None
      try:
         self.firewallSystem.UpdateDefaultPolicy(updatePolicy)
      except Exception as e:
         error = e

      self.assertTrue(error == None,
                      "error should be None")

      newPolicy = self.firewallSystem.firewallInfo.defaultPolicy
      newBlocked = newPolicy.incomingBlocked
      self.assertTrue(blocked != newBlocked,
                      "default policy should be updated")

      # Invalid argument exception
      updatePolicy.incomingBlocked = True
      updatePolicy.outgoingBlocked = False

      try:
         self.firewallSystem.UpdateDefaultPolicy(updatePolicy)
      except Exception as e:
         error = e

      self.assertTrue(error != None,
                      "error should not be None")
      self.assertTrue(isinstance(error, Vmodl.Fault.InvalidArgument),
                      "should get invalid argument fault")

      print("Test Firewall UpdateDefaultPolicy: PASS")

   def GetRulesetStatus(self, rulesetId):
      """ Return ruleset status """
      rulesets = self.firewallSystem.firewallInfo.ruleset
      for ruleset in rulesets:
         if ruleset.key == rulesetId:
            return ruleset.enabled
      return False

   def TestUpdateRulesetStatus(self):
      """
      Try to enable/disable firewall rulesets
      """
      print("Test Firewall EnableRuleset/DisableRuleset API")

      self.assertTrue(self.firewallSystem != None,
                      "firewall system should not be None")

      rulesets = self.firewallSystem.firewallInfo.ruleset

      self.assertTrue(rulesets != None,
                      "firewall system rulesets not be None")

      found = False
      # Enable/Disable sshClient ruleset
      for ruleset in rulesets:
         if ruleset.key == "sshClient":
            found = True
            enabled = ruleset.enabled
            if enabled:
               self.firewallSystem.DisableRuleset("sshClient")
               checkEnabled = self.GetRulesetStatus("sshClient")
               self.assertTrue(not checkEnabled,
                              "Disable ruleset sshClient failed")
               self.firewallSystem.EnableRuleset("sshClient")
               checkEnabled = self.GetRulesetStatus("sshClient")
               self.assertTrue(checkEnabled,
                              "Enable ruleset sshClient failed")
            else:
               self.firewallSystem.EnableRuleset("sshClient")
               checkEnabled = self.GetRulesetStatus("sshClient")
               self.assertTrue(checkEnabled,
                              "Enable ruleset sshClient failed")
               self.firewallSystem.DisableRuleset("sshClient")
               checkEnabled = self.GetRulesetStatus("sshClient")
               self.assertTrue(not checkEnabled,
                              "Disable ruleset sshClient failed")
            break

      self.assertTrue(found,
                     "Not found sshClient ruleset")
      print("Test Firewall EnableRuleset/DisableRuleset API: PASS")

   def IsAllowedHostsChanged(self, oldIpList, newIpList):
      """check whether oldIpList and newIpList is different,
         type of input parameters is Vim.Host.Ruleset.IpList
      """
      if (oldIpList != None and newIpList == None) or \
         (oldIpList == None and newIpList != None):
         return True
      if oldIpList != None and newIpList != None:
         if newIpList.allIp != oldIpList.allIp:
            return True
         ipList = newIpList.ipAddress
         netList = newIpList.ipNetwork
         if len(ipList) != len(oldIpList.ipAddress) or \
            len(netList) != len(oldIpList.ipNetwork):
            return True
         for oldip in oldIpList.ipAddress:
            notfound = True
            for newip in ipList:
               if oldip == newip:
                  notfound = False
                  break
            if notfound:
               return True
         for i in range(len(ipList)):
            if ipList[i] != oldIpList.ipAddress[i]:
               return True
         for oldnet in oldIpList.ipNetwork:
            notfound = True
            for newnet in netList:
               if oldnet.network == newnet.network and\
                  oldnet.prefixLength == newnet.prefixLength:
                  notfound = False
                  break
            if notfound:
               return True
      return False

   def TestUpdateRulesetAllowedHosts(self):
      """
      Try to configure allowediplist for sshClient ruleset
      """

      print("Test Firewall UpdateRuleset API")

      self.assertTrue(self.firewallSystem != None,
                       "firewall system should not be None")

      error = None

      # Normal case
      ipAddress1 = "1.1.1.2"
      ipAddress2 = "fd3e:29a6:0a81:e478::"
      ipNetwork1 = Vim.Host.Ruleset.IpNetwork()
      ipNetwork1.network = "2.1.1.0"
      ipNetwork1.prefixLength = 24
      ipNetwork2 = Vim.Host.Ruleset.IpNetwork()
      ipNetwork2.network = "fd3e:29a6:0a81:eabc::"
      ipNetwork2.prefixLength = 64
      spec = Vim.Host.Ruleset.RulesetSpec()
      spec.allowedHosts = Vim.Host.Ruleset.IpList()
      spec.allowedHosts.ipAddress.append(ipAddress1)
      spec.allowedHosts.ipAddress.append(ipAddress2)
      spec.allowedHosts.ipNetwork.append(ipNetwork1)
      spec.allowedHosts.ipNetwork.append(ipNetwork2)
      spec.allowedHosts.allIp = False
      self.firewallSystem.UpdateRuleset("sshClient", spec)

      rulesets = self.firewallSystem.firewallInfo.ruleset
      for ruleset in rulesets:
         if ruleset.key == "sshClient":
            newIpList = ruleset.allowedHosts
            self.assertTrue(not self.IsAllowedHostsChanged(newIpList, spec.allowedHosts),
                            "Allowed Hosts should be same")

      # Not found case
      try:
         self.firewallSystem.UpdateRuleset("NonExistRuleset", spec)
      except Exception as e:
         error = e

      self.assertTrue(error != None,
                      "error should not be None")
      self.assertTrue(isinstance(error, Vim.Fault.NotFound),
                      "should get not found fault")

      # Invalid argument case
      spec.allowedHosts.ipAddress.append("1.1.1.abcd")
      try:
         self.firewallSystem.UpdateRuleset("sshClient", spec)
      except Exception as e:
         error = e

      self.assertTrue(error != None,
                      "error should not be None")
      self.assertTrue(isinstance(error, Vmodl.Fault.InvalidArgument),
                      "should get invalid argument fault")

      print("Test Firewall UpdateRuleset API: PASS\n")


def main(argv):
    """
    Test Host Firewall Management VModl API
    """
    options = get_options()
    test = TestFirewallSystem()
    test.setOptions(options)
    suite = unittest.TestSuite()
    suite.addTest(test)
    unittest.TextTestRunner(verbosity=2).run(suite)
    return 0


# Start program
if __name__ == "__main__":
    main(sys.argv[1:])
