#!/usr/bin/python

#
# Basic tests for Vim::Vm::Guest::FileManager
#
# These require the latest tools on a VM with a supported guest
#
from __future__ import print_function

from pyVmomi import Vim, Vmodl
from pyVim.helpers import Log
from pyVim import arguments
from datetime import datetime
from urlparse import urlparse
from httplib import HTTPSConnection;
import httplib, urllib;
from guestOpsUtils import *

def waitForProcessToComplete(virtualMachine, guestAuthentication, pid):
   pids = [pid]
   while True:
      Log("Waiting for process %s to complete" % pid)
      time.sleep(3)
      result  = procMgr.ListProcesses(virtualMachine, guestAuthentication, pids)
      process = result.pop()
      if process.endTime:
         Log("Process completed with exit code: %s" % process.exitCode)
         break

def startProgramEx(virtualMachine, guestAuthentication, programPath, workingDirectory, arguments):
   spec = procDef.ProgramSpec(programPath=programPath, workingDirectory=workingDirectory, arguments=arguments)
   pid  = procMgr.StartProgram(virtualMachine, guestAuthentication, spec)
   waitForProcessToComplete(virtualMachine, guestAuthentication, pid)
   return pid

def generateStringWithSpecifiedSize(size):
   data = ""
   for i in xrange(size):
      data += "0"

   return data

def testSetup():
   global baseDir

   home = "/tmp"
   try:
      result  = procMgr.ReadEnvironmentVariable(virtualMachine, guestAuth, "HOME")
   else:
      if len(result) != 1:
         raise Exception ("Failed to fetch $HOME env var")
      home = (result[0].split('='))[1]
      Log("Fetched $HOME for guestuser: " + home)

   baseDir = home + "/foundrytests/"
   try:
      Log("Cleaning up test workspace if any")
      result = fileMgr.DeleteDirectory(virtualMachine, guestAuth, baseDir, True)
   except:
      pass

   Log("Setting up test workspace")
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, baseDir, False)
   else:
      Log("Created baseDir for foundrytests: " + baseDir)

   # changeFileAttributesTests
   dirPath = baseDir + "changeFileAttributesTests/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, False)
   else:
      Log("Created: " + dirPath)

   file = dirPath + "file-0.txt"
   if guestAdminAuth != "":
      try:
         arg = "> " + file
         result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
      else:
         Log("Created: " + file)
   else:
      Log("Skipping creation of: " + file + " (needed for special test with -R & -X for guest root user & passwd)")

   file = dirPath + "file-1.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   # createTemporaryFileTests
   dirPath = baseDir + "createTemporaryFileTests/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, False)
   else:
      Log("Created: " + dirPath)

   # makeDirectoryTests
   dirPath = baseDir + "makeDirectoryTests/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, False)
   else:
      Log("Created: " + dirPath)

   # initiateFileTransferFromGuestTests
   dirPath = baseDir + "initiateFileTransferFromGuestTests/"

   emptyDirPath = dirPath + "emptyDirectory/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, emptyDirPath, True)
   else:
      Log("Created: " + dirPath)
      Log("Created: " + emptyDirPath)

   file = dirPath + "file-1.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   # initiateFileTransferToGuestTests
   dirPath = baseDir + "initiateFileTransferToGuestTests/"

   emptyDirPath = dirPath + "emptyDirectory/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, emptyDirPath, True)
   else:
      Log("Created: " + dirPath)
      Log("Created: " + emptyDirPath)

   file = dirPath + "file-1.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   # moveFileTests
   dirPath = baseDir + "moveFileTests/"

   dirPath1 = dirPath + "dir-1/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath1, True)
   else:
      Log("Created: " + dirPath)
      Log("Created: " + dirPath1)

   symdir = dirPath + "symlink-dir-1"
   try:
      arg = "-s " + dirPath1 + " " + symdir
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ln", baseDir, arg)
   else:
      Log("Created: " + symdir)

   file = dirPath1 + "file-2.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   dirPath2 = dirPath + "dir-2/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath2, True)
   else:
      Log("Created: " + dirPath2)

   file = dirPath + "file-1.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   symfile = dirPath + "symlink-file-1"
   try:
      arg = "-s " + file + " " + symfile
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ln", baseDir, arg)
   else:
      Log("Created: " + symfile)

   file = dirPath + "file-2"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   file = dirPath + "file"
   symfile = dirPath + "symlink-nonExisting-file"
   try:
      arg = "-s " + file + " " + symfile
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ln", baseDir, arg)
   else:
      Log("Created: " + symfile)

   # deleteFileTests
   dirPath = baseDir + "deleteFileTests/"
   nonEmptyDirPath = dirPath + "nonEmptyDirectory/"
   childDirPath = nonEmptyDirPath + "childDirectory-1/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, childDirPath, True)
   else:
      Log("Created: " + dirPath)
      Log("Created: " + nonEmptyDirPath)
      Log("Created: " + childDirPath)

   symdir = dirPath + "symlink-directory"
   try:
      arg = "-s " + nonEmptyDirPath + " " + symdir
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ln", baseDir, arg)
   else:
      Log("Created: " + symdir)

   emptyDirPath = dirPath + "emptyDirectory-1/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, emptyDirPath, True)
   else:
      Log("Created: " + emptyDirPath)

   emptyDirPath = dirPath + "emptyDirectory-2/"
   try:
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, emptyDirPath, True)
   else:
      Log("Created: " + emptyDirPath)

   file = dirPath + "file-1.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   symfile = dirPath + "symlink-file-1"
   try:
      arg = "-s " + file + " " + symfile
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ln", baseDir, arg)
   else:
      Log("Created: " + symfile)

   file = dirPath + "file-2.txt"
   try:
      arg = "> " + file
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
   else:
      Log("Created: " + file)

   symfile = dirPath + "symlink-file-2"
   try:
      arg = "-s " + file + " " + symfile
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ln", baseDir, arg)
   else:
      Log("Created: " + symfile)

def testMakeDirectory():
   Log("Testing MakeDirectory")
   testDir = baseDir + "makeDirectoryTests/"

   try:
      Log("MakeDirectoryTest - 1")
      dirPath = testDir
      recursive = False
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, recursive)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      Log("MakeDirectoryTest - 2")
      dirPath = testDir + "dir-1/dir-2"
      recursive = True
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, recursive)
   else:
      Log("Test case passed")

   try:
      Log("MakeDirectoryTest - 3")
      dirPath = testDir + "dir-4"
      recursive = False
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, recursive)
   else:
      Log("Test case passed")

   try:
      Log("MakeDirectoryTest - 4")
      dirPath = testDir + "non-existing-dir-1/non-existing-dir-2"
      recursive = False
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, recursive)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      Log("MakeDirectoryTest - 5")
      baseName = ''
      for num in xrange(10):
         baseName += "abcdefghijklmnopqrstuvwxyz0123456789"
      dirPath = testDir + baseName
      recursive = False
      result = fileMgr.MakeDirectory(virtualMachine, guestAuth, dirPath, recursive)
   except Vim.Fault.FileNameTooLong as e:
      Log("Test case passed. Expected FileNameTooLong fault is thrown")
   else:
      Log("Test case failed")

def testDeleteFile():
   Log("Testing DeleteFile")
   testDir = baseDir + "deleteFileTests/"

   try:
      Log("DeleteFileTest - 1")
      dirPath = testDir + "file-1.txt"
      result = fileMgr.DeleteFile(virtualMachine, guestAuth, dirPath)
   else:
      Log("Test case passed")

   try:
      Log("DeleteFileTest - 2")
      dirPath = testDir + "non-existing-file.txt"
      result = fileMgr.DeleteFile(virtualMachine, guestAuth, dirPath)
   else:
      Log("Test case failed")

   try:
      Log("DeleteFileTest - 3")
      dirPath = testDir + "emptyDirectory-1"
      result = fileMgr.DeleteFile(virtualMachine, guestAuth, dirPath)
   except Vim.Fault.NotAFile as e:
      Log("Test case passed. Expected NotAFile fault is thrown")
   else:
      Log("Test case failed")

def testDeleteDirectory():
   Log("Testing DeleteDirectory")
   testDir = baseDir + "deleteFileTests/"

   try:
      Log("DeleteDirectoryTest - 1")
      dirPath = testDir + "file-2.txt"
      result = fileMgr.DeleteDirectory(virtualMachine, guestAuth, dirPath, False)
   except Vim.Fault.NotADirectory as e:
      Log("Test case passed. Expected NotADirectory fault is thrown")
   else:
      Log("Test case failed")

   try:
      Log("DeleteDirectoryTest - 2")
      dirPath = testDir + "non-existing-file.txt"
      result = fileMgr.DeleteDirectory(virtualMachine, guestAuth, dirPath, False)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      Log("DeleteDirectoryTest - 3")
      dirPath = testDir + "emptyDirectory-1"
      result = fileMgr.DeleteDirectory(virtualMachine, guestAuth, dirPath, False)
   else:
      Log("Test case passed")

   try:
      Log("DeleteDirectoryTest - 4")
      dirPath = testDir + "nonEmptyDirectory"
      result = fileMgr.DeleteDirectory(virtualMachine, guestAuth, dirPath, False)
   except Vim.Fault.DirectoryNotEmpty as e:
      Log("Test case passed. Expected DirectoryNotEmpty fault is thrown")
   else:
      Log("Test case failed")

   try:
      Log("DeleteDirectoryTest - 5")
      dirPath = testDir + "nonEmptyDirectory"
      result = fileMgr.DeleteDirectory(virtualMachine, guestAuth, dirPath, True)
   else:
      Log("Test case passed")

def testCreateTempFile():
   Log("Testing CreateTemporaryFile")
   testDir = baseDir + "createTemporaryFileTests"

   try:
      Log("CreateTempFile -1")
      result = fileMgr.CreateTemporaryFile(virtualMachine, guestAuth, "pre", "suf", "nonEmptyDir")
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      Log("CreateTempFile -2")
      result = fileMgr.CreateTemporaryFile(virtualMachine, guestAuth, "pre", "suf", "")
   else:
      Log(result + " Expected (A valid file path)")

   try:
      Log("CreateTempFile -3")
      result = fileMgr.CreateTemporaryFile(virtualMachine, guestAuth, "pre", "suf", testDir)
   else:
      Log(result + " Expected (A valid file path)")

   try:
      Log("CreateTempFile -4")
      result = fileMgr.CreateTemporaryDirectory(virtualMachine, guestAuth, "pre", "suf")
   else:
      Log(result + " Expected (A valid directory path)")

   try:
      Log("CreateTempFile -5")
      result = fileMgr.CreateTemporaryDirectory(virtualMachine, guestAuth, "pre", "suf", "")
   else:
      Log(result + " Expected (A valid directory path)")

   try:
      Log("CreateTempFile -6")
      result = fileMgr.CreateTemporaryFile(virtualMachine, guestAuth, "pre", "suf")
   else:
      Log(result + " Expected (A valid file path)")

def testMoveFile():
   Log("Testing MoveFile")
   testDir = baseDir + "moveFileTests/"

   try:
      srcFile = testDir + "file-1.txt"
      dstFile = testDir + "dir-1"
      Log("MoveFile - 1")
      result = fileMgr.MoveFile(virtualMachine, guestAuth, srcFile, dstFile, True)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      srcFile = testDir + "file-1.txt"
      dstFile = testDir + "dir-1/file-2.txt"
      Log("MoveFile - 2")
      result = fileMgr.MoveFile(virtualMachine, guestAuth, srcFile, dstFile, False)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      srcFile = testDir + "file-1.txt"
      dstFile = testDir + "dir-1/file-1.txt"
      Log("MoveFile - 3")
      result = fileMgr.MoveFile(virtualMachine, guestAuth, srcFile, dstFile, False)
   else:
      Log("Test case passed")

   try:
      srcFile = testDir + "dir-1/file-1.txt"
      dstFile = testDir + "dir-1/file-2.txt"
      Log("MoveFile - 4")
      result = fileMgr.MoveFile(virtualMachine, guestAuth, srcFile, dstFile, True)
   else:
      Log("Test case passed")

   try:
      srcFile = "nonExistingFile"
      dstFile = "nonExistingFile2"
      Log("MoveFile - 5")
      result = fileMgr.MoveFile(virtualMachine, guestAuth, srcFile, dstFile, True)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

def testMoveDirectory():
   testDir = baseDir + "moveFileTests/"

   try:
      srcFile = "nonExistingFile"
      dstFile = "nonExistingFile2"
      Log("MoveDirectory - 1")
      result = fileMgr.MoveDirectory(virtualMachine, guestAuth, srcFile, dstFile)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      srcFile = testDir + "dir-2"
      dstFile = testDir + "dir-1"
      Log("MoveDirectory - 2")
      result = fileMgr.MoveDirectory(virtualMachine, guestAuth, srcFile, dstFile)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      srcFile = testDir + "dir-2"
      dstFile = testDir + "dir-1/dir-2"
      Log("MoveDirectory - 3")
      result = fileMgr.MoveDirectory(virtualMachine, guestAuth, srcFile, dstFile)
   else:
      Log("Test case passed")

def testChangeFileAttributes():
   testDir = baseDir + "changeFileAttributesTests/"

   fileDate = datetime(2009, 01, 01, 01, 01, 01)
   if guestAdminAuth != "":
      fileAttributes = fileDef.PosixFileAttributes()
      fileAttributes.accessTime = fileDate
      fileAttributes.modificationTime = fileDate
      fileAttributes.ownerId = 1
      fileAttributes.groupId = 1
      fileAttributes.permissions = 484
      try:
         srcFile = testDir + "file-0.txt"
         Log("ChangeFileAttributes - 0")
         result = fileMgr.ChangeFileAttributes(virtualMachine, guestAdminAuth, srcFile, fileAttributes)
      else:
         Log("Test case passed")
   else:
      Log("Skipping test: ChangeFileAttributes - 0 (needs -R & -X for guest root user & passwd)")

   fileAttributes = fileDef.PosixFileAttributes()
   fileAttributes.accessTime = fileDate
   fileAttributes.modificationTime = fileDate
   fileAttributes.permissions = 484
   try:
      srcFile = testDir + "file-1.txt"
      Log("ChangeFileAttributes - 1")
      result = fileMgr.ChangeFileAttributes(virtualMachine, guestAuth, srcFile, fileAttributes)
   else:
      Log("Test case passed")

   fileAttributes = fileDef.WindowsFileAttributes()
   fileAttributes.hidden = False
   fileAttributes.accessTime = fileDate
   fileAttributes.modificationTime = fileDate
   try:
      srcFile = testDir + "file-1.txt"
      Log("ChangeFileAttributes - 2")
      result = fileMgr.ChangeFileAttributes(virtualMachine, guestAuth, srcFile, fileAttributes)
      # This api will pass if executed against a windows VM.
   except Vmodl.Fault.InvalidArgument as e:
      Log("Expected InvalidArgument exception is thrown")
   else:
      Log("Test case failed")

   fileAttributes = fileDef.FileAttributes()
   try:
      srcFile = testDir + "file-1.txt"
      Log("ChangeFileAttributes - 3")
      result = fileMgr.ChangeFileAttributes(virtualMachine, guestAuth, srcFile, fileAttributes)
   else:
      Log("Test case passed")

   fileAttributes = fileDef.PosixFileAttributes()
   fileAttributes.permissions = 0777

   try:
      srcFile = testDir + "file-1.txt"
      Log("ChangeFileAttributes - 4")
      result = fileMgr.ChangeFileAttributes(virtualMachine, guestAuth, srcFile, fileAttributes)
   else:
      Log("Test case passed")

   fileAttributes = fileDef.PosixFileAttributes()

   try:
      srcFile = testDir + "file-1.txt"
      Log("ChangeFileAttributes - 5")
      result = fileMgr.ChangeFileAttributes(virtualMachine, guestAuth, srcFile, fileAttributes)
   else:
      Log("Test case passed")

def uploadFile(virtualMachine, guestAuthentication, dstPath, fileAttributes, size, overwrite):
      result = fileMgr.InitiateFileTransferToGuest(virtualMachine, guestAuthentication, dstPath, fileAttributes, size, overwrite)
      url = result.replace('*', host1)
      Log("URL: %s" % url)

      urloptions = urlparse(url)
      Log("Getting the HTTPS Connection")
      conn = HTTPSConnection(urloptions.netloc)

      headers = {"Content-Length" : str(size)}
      body = generateStringWithSpecifiedSize(size)
      Log("Sending a PUT request")
      conn.request("PUT", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for PUT request: %s" % res.status)
      if res.status == 200:
         Log("Successful error code returned for PUT Request")
      else:
         Log("PUT request failed with errorcode : %s" % res.status)
         Log("Response returned for PUT request : %s" % res.read())
         raise Exception()

def testListFiles():
   # positive tests
   try:
      Log("Testing ListFiles of /tmp")
      result  = fileMgr.ListFiles(virtualMachine, guestAuth, "/tmp")
   else:
      Log("Test case passed: %s" % result)

   try:
      Log("Testing ListFiles of /bin/ls")
      result  = fileMgr.ListFiles(virtualMachine, guestAuth, "/bin/ls")
   else:
      Log("Test case passed: %s" % result)

   try:
      Log("Testing ListFiles of /tmp/*.log")
      result  = fileMgr.ListFiles(virtualMachine, guestAuth, "/tmp", matchPattern=".*.log")
   else:
      print(result)
      Log("Test case passed: %s" % result)

   # neg test
   try:
      Log("Negative test: Testing ListFiles of /tmp/ls")
      result  = fileMgr.ListFiles(virtualMachine, guestAuth, "/tmp/ls")
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

#
# Starts the src file size 0, grows it after xfer has started.
# Simulate PR 1203102
#
def testCopyFromGuestZeroGrowing():
   testDir = baseDir + "initiateFileTransferFromGuestTests/"
   srcFile = testDir + "growfile.txt"
   Log("Copying a growing file that starts size-0.")
   Log("Making initial size-0 file")
   arg = srcFile
   result  = startProgramEx(virtualMachine, guestAuth, "/bin/touch", baseDir, arg)
   try:
      body = ""
      headers = {}
      Log("InitiateFileTransferFromGuest - 7")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
      Log("Test case passed. File Transfer URL= %s" % result.url)
      url = result.url.replace('*', host1)
      Log("New URL: %s" % url)

      Log("Grow the file inside the guest.")
      arg = ">> " + srcFile
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
      Log("Grew the contents of the file: " + srcFile)

      urloptions = urlparse(url)
      Log("Getting the HTTPS Connection")
      conn = HTTPSConnection(urloptions.netloc)

      Log("Sending a GET request")
      conn.request("GET", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for GET request: %s" % res.status)
      Log("Response body for GET request: %s" % res.read())
      if res.status == 200:
         Log("Successful error code returned for GET Request")
         Log("HTTP Content (Partial) = " + res.read())
      else:
         raise Exception()
   except Exception:
      Log("Test case failed")
      raise


def testInitiateFileTransferToGuest():
   testDir = baseDir + "initiateFileTransferToGuestTests/"
   srcFile = testDir + "file-1.txt"
   fileDate = datetime(2009, 01, 01, 01, 01, 01)
   fileAttributes = fileDef.PosixFileAttributes()
   fileAttributes.accessTime = fileDate
   fileAttributes.modificationTime = fileDate
   fileAttributes.ownerId = 1
   fileAttributes.groupId = 1
   fileAttributes.permissions = 484
   try:
      Log("InitiateFileTransferToGuest - 1")
      result = fileMgr.InitiateFileTransferToGuest(virtualMachine, guestAuth, srcFile, fileAttributes, 100, False)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      Log("InitiateFileTransferToGuest - 2")
      result = fileMgr.InitiateFileTransferToGuest(virtualMachine, guestAuth, srcFile, fileAttributes, 100, True)
   else:
      Log("URL for the file transfer: %s" %result)

   try:
      srcFile = testDir
      Log("InitiateFileTransferToGuest - 3")
      result = fileMgr.InitiateFileTransferToGuest(virtualMachine, guestAuth, srcFile, fileAttributes, 100, True)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      srcFile = testDir + "emptyDirectory"
      Log("InitiateFileTransferToGuest - 4")
      result = fileMgr.InitiateFileTransferToGuest(virtualMachine, guestAuth, srcFile, fileAttributes, 100, True)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   try:
      fileAttributes = fileDef.WindowsFileAttributes()
      fileAttributes.createTime = fileDate
      fileAttributes.accessTime = fileDate
      fileAttributes.modificationTime = fileDate
      fileAttributes.hidden = True
      srcFile = testDir + "emptyDirectory"
      Log("InitiateFileTransferToGuest - 5")
      result = fileMgr.InitiateFileTransferToGuest(virtualMachine, guestAuth, srcFile, fileAttributes, 100, True)
      # This api will pass if executed against a windows VM.
   except Vmodl.Fault.InvalidArgument as e:
      Log("Expected InvalidArgument exception is thrown")
   else:
      Log("Test case failed")

   try:
      fileAttributes = fileDef.FileAttributes()
      dstPath = testDir + "file-1.txt"
      Log("InitiateFileTransferToGuest - 6")
      uploadFile(virtualMachine, guestAuth, dstPath, fileAttributes, 100, True)
   else:
      Log("Test case passed")

   try:
      fileAttributes = fileDef.PosixFileAttributes()
      fileAttributes.ownerId = 1
      fileAttributes.permissions = 0444
      dstPath = testDir + "file-2.txt"
      Log("InitiateFileTransferToGuest - 7")
      uploadFile(virtualMachine, guestAuth, dstPath, fileAttributes, 100, True)
   else:
      Log("Test case passed")

def testInitiateFileTransferFromGuest():
   testDir = baseDir + "initiateFileTransferFromGuestTests/"

   srcFile = baseDir
   try:
      Log("InitiateFileTransferFromGuest - 1")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   srcFile = testDir + "nonexistingfile"
   try:
      Log("InitiateFileTransferFromGuest - 2")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   srcFile = testDir + "emptyDirectory"
   try:
      print("InitiateFileTransferFromGuest - 3")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
   except Exception:
      Log("Test case passed")
   else:
      Log("Test case failed")

   srcFile = testDir + "file-1.txt"
   try:
      body = ""
      headers = {}
      Log("InitiateFileTransferFromGuest - 4")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
      Log("Test case passed. File Transfer URL= %s" % result.url)
      url = result.url.replace('*', host1)
      Log("New URL: %s" % url)

      urloptions = urlparse(url)
      Log("Getting the HTTPS Connection")
      conn = HTTPSConnection(urloptions.netloc)

      Log("Sending a GET request")
      conn.request("GET", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for GET request: %s" % res.status)
      if res.status == 200:
         Log("Successful error code returned for GET Request")
         Log("HTTP Content (Partial) = " + res.read())
      else:
         raise Exception()

      Log("Sending a GET request second time")
      conn = HTTPSConnection(urloptions.netloc)
      conn.request("GET", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for GET request: %s" % res.status)
      if res.status == 404:
         Log("Successful error code returned for second GET Request")
      else:
         raise Exception()
   except Exception:
      Log("Test case failed")
      raise

   srcFile = testDir + "file-1.txt"
   try:
      body = ""
      headers = {}
      Log("InitiateFileTransferFromGuest - 5")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
      Log("Test case passed. File Transfer URL= %s" % result.url)
      url = result.url.replace('*', host1)
      Log("New URL: %s" % url)

      Log("Expanding the file inside the guest.")
      arg = ">> " + srcFile
      result = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
      Log("Expanded the content in the file: " + srcFile)

      urloptions = urlparse(url)
      Log("Getting the HTTPS Connection")
      conn = HTTPSConnection(urloptions.netloc)

      Log("Sending a GET request")
      conn.request("GET", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for GET request: %s" % res.status)
      if res.status == 200:
         Log("Successful error code returned for GET Request")
         Log("HTTP Content (Partial) = " + res.read())
      else:
         raise Exception()
   except Exception:
      Log("Test case failed")
      raise

   srcFile = testDir + "file-1.txt"
   try:
      body = ""
      headers = {}
      Log("InitiateFileTransferFromGuest - 6")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
      Log("Test case passed. File Transfer URL= %s" % result.url)
      url = result.url.replace('*', host1)
      Log("New URL: %s" % url)

      Log("Truncating the file inside the guest.")
      arg = "> " + srcFile
      result  = startProgramEx(virtualMachine, guestAuth, "/bin/ls", baseDir, arg)
      Log("Truncated the content in the file: " + srcFile)

      urloptions = urlparse(url)
      Log("Getting the HTTPS Connection")
      conn = HTTPSConnection(urloptions.netloc)

      Log("Sending a GET request")
      conn.request("GET", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for GET request: %s" % res.status)
      if res.status == 200:
         Log("Successful error code returned for GET Request")
         Log("HTTP Content (Partial) = " + res.read())
      else:
         raise Exception()
   except Exception:
      Log("Test case failed")
      raise

   testCopyFromGuestZeroGrowing()

   # this powers off the VM, so do it last
   srcFile = testDir + "file-1.txt"
   try:
      body = ""
      headers = {}
      Log("InitiateFileTransferFromGuest - 8")
      result = fileMgr.InitiateFileTransferFromGuest(virtualMachine, guestAuth, srcFile)
      Log("Test case passed. File Transfer URL= %s" % result.url)
      url = result.url.replace('*', host1)
      urloptions = urlparse(url)

      # power off the VM if needed
      powerOffVM(virtualMachine)

      Log("Getting the HTTPS Connection")
      conn = HTTPSConnection(urloptions.netloc)

      Log("Sending a GET request after the VM is powered off")
      conn.request("GET", urloptions.path + "?" + urloptions.query, body, headers)
      res = conn.getresponse()
      Log("Response code for GET request: %s" % res.status)
      if res.status == 404:
         Log("Successful error code returned for GET Request")
      else:
         raise Exception()

      powerOnVM(virtualMachine)

   except Exception:
      Log("Test case failed")
      raise


def main():

   # Process command line
   options = get_options()

   global svcInst
   global virtualMachine
   global guestAdminAuth
   global guestAuth
   global guestAuthBad
   [svcInst, virtualMachine, guestAdminAuth,
    guestAuth, guestAuthBad] = init(options.host, options.user, options.password,
                                    options.vmname, options.vmxpath, options.guestuser,
                                    options.guestpassword, options.guestrootuser,
                                    options.guestrootpassword)

   # get the processManager object
   global procMgr
   procMgr = svcInst.content.guestOperationsManager.processManager
   global procDef
   procDef = Vim.Vm.Guest.ProcessManager

   # get the fileManager object
   global fileMgr
   fileMgr = svcInst.content.guestOperationsManager.fileManager
   global fileDef
   fileDef = Vim.Vm.Guest.FileManager

   global host1
   host1 = options.host

   testNotReady(procMgr, virtualMachine, guestAuth)
   waitForTools(virtualMachine)

   testSetup()

   testMakeDirectory()
   testDeleteFile()
   testDeleteDirectory()
   testCreateTempFile()
   testMoveFile()
   testMoveDirectory()
   testChangeFileAttributes()
   testListFiles()
   testInitiateFileTransferToGuest()

   # IMPORTANT: This test powers the VM on and off
   # This may delete the workspace.
   # So keep this as the last test.
   testInitiateFileTransferFromGuest()

# Start program
if __name__ == "__main__":
   main()
   Log("fileManager tests completed")
