#!/usr/bin/python

from __future__ import print_function

import sys
import os
from pyVim.connect import Connect, Disconnect
from pyVmomi import Vim,VmomiSupport
from pyVim import vm, host
from pyVim import arguments
import atexit


def CreateLocalDs(dsName, rootPath):
    try:
        print("Creating local datastore " + dsName + " ...")
        localDs = datastoreSystem.CreateLocalDatastore(dsName,
                                                       rootPath)
        if localDs == None or localDs.GetInfo().GetName() != dsName:
            print("Failed to create datastore")
            return None
        else:
            print("Successfully created datastore " + dsName)
            return localDs
    except Exception as e:
        print("Failed to create datastore %s" % e)
        return None


def main():
    supportedArgs = [ (["h:", "host="], "localhost", "Host name", "host"),
                      (["u:", "user="], "root", "User name", "user"),
                      (["p:", "pwd="], "ca$hc0w", "Password", "pwd"),
                      (["d:", "dsName="], "testDs", "Name of the local datastore", "dsName"),
                      (["r:", "rootPath="], "/tmp", "Local directory for use by this script", "rootPath"),
                      (["n:", "newName="], "foo", "Alternative name for the local datastore", "newName"),
                      (["i:", "numiter="], "1", "Number of iterations", "iter") ]

    supportedToggles = [ (["usage", "help"], False, "Show usage information", "usage")]

    args = arguments.Arguments(sys.argv, supportedArgs, supportedToggles)
    if args.GetKeyValue("usage") == True:
        args.Usage()
        sys.exit(0)

    # Process command line
    dsName = args.GetKeyValue("dsName")
    numiter = int(args.GetKeyValue("iter"))
    rootPath = args.GetKeyValue("rootPath")
    newName = args.GetKeyValue("newName")
    mkdircmd = "mkdir -p "
    rmcmd = "rm -rf "
    rootPath.rstrip('/')

    # Connect
    si = Connect(host=args.GetKeyValue("host"),
                 user=args.GetKeyValue("user"),
                 pwd=args.GetKeyValue("pwd"))
    atexit.register(Disconnect, si)

    hostSystem = host.GetHostSystem(si)
    hostConfigManager = hostSystem.GetConfigManager()
    global datastoreSystem
    datastoreSystem = hostConfigManager.GetDatastoreSystem()

    i = 0
    while i < numiter:
        print("Current Datastore list: ")
        datastoreList = hostSystem.GetDatastore()
        print(datastoreList)

        print("Running Positive tests for Local Datastores")
        try:
            # Add a local datastore
            dsPath1 = rootPath + "/testDs1"
            os.system(mkdircmd + dsPath1)
            localDs = CreateLocalDs(dsName, dsPath1)
            if localDs == None:
                return

            # Print datastore information
            dsInfo = localDs.GetInfo()
            print("Datastore info:")
            print(dsInfo)
            print("Datastore summary:")
            print(localDs.GetSummary())
            dsPath = dsInfo.GetUrl()

            # Rename the datastore
            print("Renaming the datastore to " + newName)
            localDs.Rename(newName)
            if localDs.GetInfo().GetName() != newName:
                print("Failed to rename datastore")
                return
            print("Datastore summary after rename: ")
            print(localDs.GetSummary())

            # Refresh the datastore after removing backing
            print("Removing directory " + dsPath1+ "...")
            os.system(rmcmd + dsPath1)
            print("Refreshing datastore")
            localDs.Refresh()
            if localDs.GetSummary().GetAccessible() != False:
                print("Failed to refresh datastore state as inaccessible")
                return
            print("Refreshed Datastore state as inaccesible")
            print(localDs.GetSummary())

            # Remove the datastore
            print("Removing datastore...")
            localDs.Destroy()
            datastoreList = hostSystem.GetDatastore()
            for dsRef in datastoreList:
                if dsRef.GetInfo().GetUrl() == dsPath:
                    print("Failed to remove datastore")
                    return
            print("Datastore list after removal: ")
            print(datastoreList)

        except Exception as e:
            print("Failed datastore op: %s" % e)
            return

        print("Positive tests for local datastores: PASS")
        print("Negative tests for local datastores")

        # Add datastore with non-existant backing
        dsPath2 = rootPath + "/invalidBacking"
        if CreateLocalDs(dsName, dsPath2) != None:
            print("Created datastore with invalid backing")
            return
        print("Datastore with invalid backing was rejected")

        # Add a datastore with a path clash
        os.system(mkdircmd + dsPath1)
        localDs = CreateLocalDs(dsName, dsPath1)
        if localDs == None:
            return
        if CreateLocalDs(newName, dsPath1) != None:
            print("Created datastore with clashing path")
            return
        print("Datastore with clashing path was rejected")

        # Add a nested datastore
        dsPath2 = dsPath1 + "/testDs2"
        os.system(mkdircmd + dsPath2)
        if CreateLocalDs(newName, dsPath2) != None:
            print("Created datastore with nested directory path")
            return
        print("Datastore with nested path was rejected")


        # Add a datastore with a name clash
        dsPath3 = rootPath + "/testDs2"
        os.system(mkdircmd + dsPath3)
        if CreateLocalDs(dsName, dsPath3) != None:
            print("Created datastore with clashing name")
            return
        print("Datastore with clashing name was rejected")

        # Cleanup
        localDs.Destroy()
        os.system(rmcmd + dsPath3)
        os.system(rmcmd + dsPath1)
        print("Negative tests for local datastores: PASS")
        i = i+1


# Start program
if __name__ == "__main__":
    main()

