#!/usr/bin/env python

__doc__ = """
-- patchManagerTests.py --
   This module test PatchManger related functionality.

   Usage:
   ../py.sh patchManagerTests.py -H <host name/ip> -u <user name> -p <password>
"""

import sys
import logging
import unittest
from pyVim.connect import Connect, Disconnect
from pyVmomi import Vim, Vmodl
from optparse import OptionParser
from pyVim import host
from pyVim.task import WaitForTask

MetaUrls = [
    "http://pdp-server14.eng.vmware.com/depot/TestVibsDepot/4/metadata12.zip",
    ]

# "http://engweb.vmware.com/~imourza/depot/vmware/vmware-cos-metadata.zip",

BundleUrls = [
    "/exit26/home/evan/vibs/thirdparty/swordfish-071408.zip",
    ]
VibUrls = [
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-01.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-02.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-03.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-04.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-05.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-06.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-07.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-08.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-09.vib",
    "http://engweb.vmware.com/~imourza/depot/vmware/classic/ESX40-SIMPLE-10.vib",
    ]

FAILED_VERSION_NUMBER = "-0.0"

def get_options():
    """
    Supports the command-line arguments listed below
    """

    parser = OptionParser()
    parser.add_option("-H", "--host",
                      default="localhost",
                      help="remote host to connect to")
    parser.add_option("-u", "--user",
                      default="root",
                      help="User name to use when connecting to hostd")
    parser.add_option("-p", "--password",
                      default="",
                      help="Password to use when connecting to hostd")
    (options, _) = parser.parse_args()
    return options


class TestPatchManager(unittest.TestCase):

    def setOptions(self, options):
        """
        Command line options
        """
        self.options = options

    def setUp(self):
        """
        Setting test suite
        """
        options = get_options()

        self.si = Connect(host=self.options.host,
                        user=self.options.user,
                        version="vim.version.version9",
                        pwd=self.options.password)
        content = self.si.RetrieveContent()
        rootFolder = content.GetRootFolder()
        dataCenter = rootFolder.GetChildEntity()[0]
        hostFolder = dataCenter.hostFolder
        host = hostFolder.childEntity[0]
        self.hostSystem = host.host[0]
        configManager = self.hostSystem.GetConfigManager()
        self._patchManager = configManager.GetPatchManager()

    def tearDown(self):
        """
        Reset test suite
        """
        Disconnect(self.si)

    def runTest(self):
        self.TestCheck()
        self.TestScanV2()
        self.TestStage()
        self.TestInstallV2()

    def InvokeAndTrack(self,
                       func,
                       *args,
                       **kw):
        self._result = None
        task = func(*args, **kw)
        # self.failIfRaises(
        #    "Waiting for task expected not to raise.",
        WaitForTask(task)
        info = task.GetInfo()
        if info is not None:
            self._result = info.GetResult()

    def TestCheck(self):
        """
        Verify Check operation.
        """
        self.InvokeAndTrack(
            self._patchManager.Check,
            MetaUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestCheckOfflineBundle(self):
        """
        Verify Check operation.
        """
        self.InvokeAndTrack(
            self._patchManager.Check,
            None, BundleUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestScanV2(self):
        """
        Verify ScanV2 operation.
        """
        self.InvokeAndTrack(
            self._patchManager.ScanV2,
            MetaUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestScanV2OfflineBundle(self):
        """
        Verify ScanV2 operation.
        """
        self.InvokeAndTrack(
            self._patchManager.ScanV2,
            None, BundleUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestStage(self):
        """
        Verify Stage operation.
        """
        self.InvokeAndTrack(
            self._patchManager.Stage,
            MetaUrls, None, VibUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestStageOfflineBundle(self):
        """
        Verify Stage operation.
        """
        self.InvokeAndTrack(
            self._patchManager.Stage,
            None, BundleUrls, VibUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestInstallV2(self):
        """
        Verify InstallV2 operation.
        """
        self.InvokeAndTrack(
            self._patchManager.InstallV2,
            MetaUrls, None, VibUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestInstallV2OfflineBundle(self):
        """
        Verify InstallV2 operation.
        """
        self.InvokeAndTrack(
            self._patchManager.InstallV2,
            None, BundleUrls, VibUrls)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestInstallV2NoSigCheck(self):
        """
        Verify InstallV2 operation.
        """
        spec = self._patchManager.PatchManagerOperationSpec()
        spec.cmdOption = "--nosigcheck"
        self.InvokeAndTrack(
            self._patchManager.InstallV2,
            MetaUrls, None, VibUrls, spec)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestInstallV2NoSigCheckAndMaintenanceMode(self):
        """
        Verify InstallV2 operation.
        """
        spec = self._patchManager.PatchManagerOperationSpec()
        spec.cmdOption = "--nosigcheck --maintenancemode"
        self.InvokeAndTrack(
            self._patchManager.InstallV2,
            MetaUrls, None, VibUrls, spec)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestQuery(self):
        """
        Verify Query operation.
        """
        self.InvokeAndTrack(self._patchManager.Query)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestCheckEmptyDepotUrl(self):
        """
        Check, empty depot url, should raise.
        """
        self.failUnlessRaises(
            "Check with empty meta url should raise.",
            Vmodl.Fault.InvalidArgument,
            self.InvokeAndTrack,
            self._patchManager.Check,
            ["",]
            )
        self.failUnless(self._result is None,
                        "Expected no results.")

    def TestCheckDepotUrlDoesntExist(self):
        """
        Check, bad depot url, the call should succeed.
        """
        self.InvokeAndTrack(
            self._patchManager.Check,
            ["file:///fake_location/metadata.zip",]
            )
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestScanV2EmptyDepotUrl(self):
        """
        Check, empty depot url, should raise.
        """
        self.failUnlessRaises(
            "ScanV2 with empty meta url should raise.",
            Vmodl.Fault.InvalidArgument,
            self.InvokeAndTrack,
            self._patchManager.ScanV2,
            ["",]
            )
        self.failUnless(self._result is None,
                        "Expected no results.")

    def TestScanV2DepotUrlDoesntExist(self):
        """
        Check, bad depot url, the call should succeed.
        """
        self.InvokeAndTrack(
            self._patchManager.ScanV2,
            ["file:///fake_location/metadata.zip",]
            )
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")


    def TestStageNoSig(self):
        """
        Verify Stage operation, no signature check.
        """
        spec = self._patchManager.PatchManagerOperationSpec()
        spec.cmdOption = "--nosigcheck"
        self.InvokeAndTrack(
            self._patchManager.Stage,
            MetaUrls, None, VibUrls, spec)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestStageMaintenance(self):
        """
        Verify Stage operation, maintenance mode.
        """
        spec = self._patchManager.PatchManagerOperationSpec()
        spec.cmdOption = "--maintenancemode"
        self.InvokeAndTrack(
            self._patchManager.Stage,
            MetaUrls, None, VibUrls, spec)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestStageMultOptions(self):
        """
        Verify Stage operation, multiple options.
        """
        spec = self._patchManager.PatchManagerOperationSpec()
        spec.cmdOption = "--maintenancemode --nosigcheck"
        self.InvokeAndTrack(
            self._patchManager.Stage,
            MetaUrls, None, VibUrls, spec)
        self.failIf(self._result is None,
                    "Expected results.")
        self.failIf(self._result.GetXmlResult() == "",
                    "Expected non-empty results.")
        logging.debug(self._result.GetXmlResult())
        self.failIf(self._result.GetVersion() == FAILED_VERSION_NUMBER,
                    "Expected valid version number in an xml response.")

    def TestScan(self):
        """
        Scan, pre-K/L version.
        """
        locator = self._patchManager.Locator()
        locator.url = MetaUrls[0]
        self.InvokeAndTrack(
            self._patchManager.Scan,
            locator
            )
        # TODO: Actually, it looks like results COULD be empty
        # TODO: Need to check return code?
        self.failIf(len(self._result) == 0,
                    "Expected non-empty results array.")

        r = self._result[0]
        self.failIf(r.GetId() == "",
                    "Expected non-empty ID.")
        applicable = r.GetApplicable()
        self.failUnless(applicable == True or applicable == False,
                    "applicable: Expected True or False.")
        vmOffReq = r.GetVmOffRequired()
        self.failUnless(vmOffReq == True or vmOffReq == False,
                    "vmOffRequired: Expected True or False.")
        restartReq = r.GetRestartRequired()
        self.failUnless(restartReq == True or restartReq == False,
                    "restartRequired: Expected True or False.")
        # TODO: Check results

    def TestInstall(self):
        """
        Install, pre-K/L version.
        """
        locator = self._patchManager.Locator()
        locator.url = MetaUrls[0]
        self.InvokeAndTrack(
            self._patchManager.Install,
            locator,
            VibUrls[0]
            )
def main(argv):
    """
    Test Patch Manager VModl API
    """
    options = get_options()
    test = TestPatchManager()
    test.setOptions(options)
    suite = unittest.TestSuite()
    suite.addTest(test)
    unittest.TextTestRunner(verbosity=2).run(suite)


# Start program
if __name__ == "__main__":
    main(sys.argv[1:])

