#!/usr/bin/python

from __future__ import print_function

import sys
from pyVmomi import Vim
from pyVmomi import VmomiSupport
from pyVim.connect import Connect, Disconnect
from pyVim.task import WaitForTask
from pyVim import folder
from pyVim import arguments
import atexit

def CheckIfPresent(svcList, testNsp):
   for svcSpec in svcList:
      if testNsp == svcSpec.GetServerNamespace():
         return True
   return False

def TestAddService(proxySvc, svcSpec):
   testNsp = svcSpec.GetServerNamespace()
   print("Adding service " + testNsp)
   proxySvc.AddService(svcSpec)
   svcList = proxySvc.GetServiceList()
   print("\nProxy Service List after addition:")
   print(svcList)
   if CheckIfPresent(svcList, testNsp) == True:
      print("Successfully added service " + testNsp)
   else:
      raise Exception("SvcSpec entry for " + testNsp + " not found!")

def TestRemoveService(proxySvc, svcSpec):
   testNsp = svcSpec.GetServerNamespace()
   print("Removing service " + testNsp)
   proxySvc.RemoveService(svcSpec)
   svcList = proxySvc.GetServiceList()
   print("\nProxy Service List after removal:")
   print(svcList)
   if CheckIfPresent(svcList, testNsp) == False:
      print("Successfully removed service " + testNsp)
   else:
      raise Exception("SvcSpec entry for " + testNsp + " found!")

def TestAddAndRemoveService(proxySvc, svcSpec):
   TestAddService(proxySvc, svcSpec)
   TestRemoveService(proxySvc, svcSpec)

def main():
   supportedArgs = [ (["h:", "host="], "localhost", "Host name", "host"),
                     (["u:", "user="], "root", "User name", "user"),
                     (["p:", "pwd="], "ca$hc0w", "Password", "pwd"),
                     (["i:", "numiter="], "1", "Number of iterations", "iter")]

   supportedToggles = [ (["usage", "help"], False, "Show usage information", "usage") ]

   args = arguments.Arguments(sys.argv, supportedArgs, supportedToggles)
   if args.GetKeyValue("usage") == True:
      args.Usage()
      sys.exit(0)

   # Connect
   si = Connect(host=args.GetKeyValue("host"),
                user=args.GetKeyValue("user"),
                pwd=args.GetKeyValue("pwd"))
   atexit.register(Disconnect, si)

   # Process command line
   numiter = int(args.GetKeyValue("iter"))
   testNsp = "/pyTest"

   # Retrieve proxy instance
   content = si.RetrieveInternalContent()
   proxySvc = content.GetProxyService()

   if proxySvc == None:
      print("ProxySvc instance is NULL!")
      sys.exit(0)

   # Query proxy config information.
   print("\nProxy Config Info\n==================")
   print("\nProxy Http Port : %s" % proxySvc.GetHttpPort())
   print("Proxy Https Port : %s" % proxySvc.GetHttpsPort())
   svcList = proxySvc.GetServiceList()
   print("Proxy Service List :")
   print(svcList)

   # Remove entries from a previous run, if present.
   try:
      svcSpec = Vim.ProxyService.ServiceSpec()
      svcSpec.SetServerNamespace(testNsp)
      svcSpec.SetAccessMode("httpAndHttps")
      proxySvc.RemoveService(svcSpec)
   except Vim.Fault.NotFound:
      print("\nService %s not present in Proxy Service List.\n" % testNsp)

   # Create ServiceSpec entries
   localSvcSpec = Vim.ProxyService.LocalServiceSpec()
   localSvcSpec.SetServerNamespace(testNsp)
   localSvcSpec.SetAccessMode("httpOnly")
   localSvcSpec.SetPort(9999)

   rmtSvcSpec = Vim.ProxyService.RemoteServiceSpec()
   rmtSvcSpec.SetServerNamespace(testNsp)
   rmtSvcSpec.SetAccessMode("httpsOnly")
   rmtSvcSpec.SetHostName("dummyhost.domain.com")
   rmtSvcSpec.SetPort(9999)

   npSvcSpec = Vim.ProxyService.NamedPipeServiceSpec()
   npSvcSpec.SetServerNamespace(testNsp)
   npSvcSpec.SetAccessMode("httpsWithRedirect")
   npSvcSpec.SetPipeName("dummy-pipe")

   for i in range(numiter):
      try:
         TestAddService(proxySvc, localSvcSpec)
         try:
            TestAddService(proxySvc, localSvcSpec)
            print("Error: Attempt to add duplicate service entries did not throw exception")
            sys.exit(0)
         except Vim.Fault.AlreadyExists:
            print("Verified exception on attempt to add duplicate service entries")
         TestRemoveService(proxySvc, localSvcSpec);
         try:
            TestRemoveService(proxySvc, localSvcSpec)
            print("Error: Attempt to remove non-existent service entry did not throw exception.")
            sys.exit(0)
         except Vim.Fault.NotFound:
            print("Verified exception on attempt to remove non-existent service entry.")
         TestAddAndRemoveService(proxySvc, rmtSvcSpec);
         TestAddAndRemoveService(proxySvc, npSvcSpec);
      except Exception as e:
         print("Caught exception %s" % e)
         sys.exit(0)

   print("\nCompleted Proxy tests.")


# Start program
if __name__ == "__main__":
    main()
