#!/usr/bin/python
#
# vmTestHbr.py
#
# Simple unit tests of hostd, primary-side HBR infrastructure support.
# Just create some VMs, register them with hostd/hbr and manipulate the
# HBR settings.
#

import sys
import random
import traceback
from pyVmomi import Vim
from pyVim import connect
from pyVim.connect import Connect, Disconnect
from pyVim import vm
from pyVim import folder
from pyVim import arguments
from pyVmomi import Vmodl
from pyVim.task import WaitForTask
from pyVim.helpers import Log
from pyVmomi import Hostd
import time
import types
import atexit

#
# Get the controller for the specified disk device
#
def GetDiskController(vmname, diskDevice):
   """
   Get the controller for the specified disk device

   @param vmname     [in] The VM the disk belongs to
   @param diskDevice [in] The disk device
   """

   devices = []

   for device in vmname.GetConfig().GetHardware().GetDevice():
      if device.GetKey() == diskDevice.GetControllerKey():
         return device

   return None


#
# Disable replication for the specified VM
#
def DisableReplication(hbrManager, vmname, errorsOkay=False):
   """
   Disable replication for the specified virtual machine

   @param vmname [in] VM to invoke operation on
   @param errorsOkay [in, optional] Wether to eat errors or let them through
   """

   try:
      task = hbrManager.DisableReplication(vmname)
      WaitForTask(task)
   except Vim.Fault.ReplicationVmConfigFault as e:
      if errorsOkay:
         Log("Got exception '%s' disabling replication, but its okay" % str(e))
      else:
         Log("FAIL: Got unexpected exception while disabling replication.")
         raise # re-raise the exception
   return


#
# Enable reeplication for the specified VM
#
def EnableReplication(hbrManager,
                      vmname,
                      destination="127.0.0.1",
                      port=1234,
                      rpo=10,
                      quiesce = False):
   """
   Enable replication for the specified virtual machine

   @param vmname      [in] VM to invoke operation on
   @param destination [in] IP address of destination
   @param rpo         [in] RPO in minutes
   @param quiesce     [in] Bool indicating whether to call tools to quiesce
   """

   repVmInfo = Vim.Vm.ReplicationConfigSpec()

   Log("Setup config spec for VM");
   repVmInfo.SetVmReplicationId(vmname.GetConfig().GetUuid())
   repVmInfo.SetDestination(destination)
   repVmInfo.SetPort(port)
   repVmInfo.SetRpo(long(rpo))
   repVmInfo.SetQuiesceGuestEnabled(quiesce)
   repVmInfo.SetOppUpdatesEnabled(False)

   devices = []
   disks = []

   Log("Setup replication for all disks in VM");
   for device in vmname.GetConfig().GetHardware().GetDevice():
      if isinstance(device, Vim.Vm.Device.VirtualDisk):
         controller = GetDiskController(vmname, device)

         if controller == None:
            raise Exception("Invalid controller for device " + str(device))

         # XXX This assumes that the VM does not have any other types of
         #     controllers, which is valid for now, but we need to add tests
         #     for the other controller types
         if not (isinstance(controller, Vim.Vm.Device.VirtualSCSIController) or
                 isinstance(controller, Vim.Vm.Device.VirtualIDEController)):
            raise Exception("Unexpected controller type " + str(controller))

         repDiskInfo = Vim.Vm.ReplicationConfigSpec.DiskSettings()
         repDiskInfo.SetKey(device.GetKey())
         repDiskInfo.SetDiskReplicationId(str(device.GetKey()))

         disks.append(repDiskInfo)

   repVmInfo.SetDisk(disks)

   Log("Configuring VM for replication")
   task = hbrManager.EnableReplication(vmname, repVmInfo)
   WaitForTask(task)

   return


#
# Print the replication configuration for the specified VM
#
def PrintConfiguration(hbrManager, vmname):
   """
   Print the replication configuration

   @param vmname [in] VM to invoke operation on
   """

   repVmInfo = hbrManager.RetrieveReplicationConfig(vmname)

   # XXX Having a better print format would be nice, but this works
   Log("Replication configuration: " + str(repVmInfo))

   return


#
# Example of a command that needs to use the VMODL interfaces
# exposed by HbrInternalSystem
#
def InternalCommand(si, vmname):
   """ XXX This is a stub to show how one might call things
       from the HbrInternalSystemInterface rather then the public
       VM interface """

   hbrInternalSystem = Hostd.HbrInternalSystem("ha-hbr-internal-manager",
                                               si._stub)



   #
   # We expect stopping offline replication to fail as the VM
   # isn't being replicated offline currently
   #
   try:
      hbrInternalSystem.StopOfflineVmReplication(vmname)
   except Exception as e:
      Log("Expected exception: Stopping offline replication failed.")
   else:
      raise Exception("Stopping offline replication should have failed!")

   return


#
# main
#
def main():
   supportedArgs = [ (["h:", "host="], "localhost", "Host name", "host"),
		     (["u:", "user="], "root", "User name", "user"),
                     (["p:", "pwd="], "", "Password", "pwd") ]

   supportedToggles = [ (["usage", "help"],
                        False,
                        "Show usage information",
                        "usage") ]

   args = arguments.Arguments(sys.argv, supportedArgs, supportedToggles)
   if args.GetKeyValue("usage") == True:
      args.Usage()
      sys.exit(0)

   host = args.GetKeyValue("host")

   status = "PASS"
   rc=0

   try:
      # Connect to hosts
      si = connect.Connect(host=host,
                           user=args.GetKeyValue("user"),
                           pwd=args.GetKeyValue("pwd"),
                           version="vim.version.version9")

      atexit.register(Disconnect, si)

      hbrManager = si.RetrieveInternalContent().hbrManager

      # If we find a VM with the name we will re-use it, the assumption
      # being that the script is the only that creates it.
      vmname = vm.CreateOrReturnExisting("vmTestHbr-private-test-vm", 2)

      Log("Removing any replication configuration for " + vmname.GetName())
      DisableReplication(hbrManager, vmname, True)

      Log("Enabling replication for " + vmname.GetName())
      EnableReplication(hbrManager, vmname)

      Log("Dumping replication configuration for" + vmname.GetName())
      PrintConfiguration(hbrManager, vmname)
      InternalCommand(si, vmname)

      # XXX We should add some more tests here that actually power-on
      # the VM, maybe extract the state, etc

      Log("Removing any replication configuration for " + vmname.GetName())
      DisableReplication(hbrManager, vmname)

      vm.Destroy(vmname)

   except Exception as e:
      (excType, excVal, excTB) = sys.exc_info()
      Log("Caught exception: ")
      traceback.print_exception(excType, excVal, excTB)
      status = "FAIL"
      rc=42

   Log("TEST RUN COMPLETE: " + status)
   return rc


# Start program
if __name__ == "__main__":
   sys.exit(main())
