#!/usr/bin/python

'''
Adapted from existing tests in vim/py/tests, e.g. qaUnitTests.py, snapshots.py
Goal: Perform snapshot related functional tests on esxi host.
Example:
Real VM -
  $VMTREE/vim/py/py.sh  $VMTREE/vim/py/tests/vmTestSnapshot.py -h "localhost" -u "root" -p "" -v "SnapshotTestRealVM"
DummyVM -
  $VMTREE/vim/py/py.sh  $VMTREE/vim/py/tests/vmTestSnapshot.py -h "localhost" -u "root" -p ""

BaseState of vmTestSnapshot:
 Note: Please make sure the target VM is OK to test with. All the existing snapshots can be removed.
 If the specified test VM cannot be found, create a dummy VM "SnapshotTestVM" to run test with.
1. Power off the VM if it is not in poweredOff state.
2. Remove all existing snapshot of current VM.

Functions to be tested:
pyVim: vm.CreateSnapshot, name, desc,
   memory, quiesce
   True, True
   True, False
   False, True
   False, False
memory True
   take snapshot VM state
     poweredOn
memory False
   take snapshot VM state
     poweredOn
     suspended
     poweredOff
   memory option is disabled if VM is in poweredOff or suspended state.
quiesce True
   The VM quiesce only apply when taking snapshot at poweredOn state. If VM is not powered on, enable quiesce make not difference. quiesce option is disabled if VM is in poweredOff or suspended state.

Baseline TestCases
Take snapshot in following VM power states:
   poweredOn - SNPoweredOn
   suspended - SNSuspened
   poweredOff - SNPoweredOff
  Todo: Before taking the snapshot: If VM is a real OS and support file operations, create Snapshot ID file in Guest OS for verification proposes. Cleanup existing snapshot ID file.
  Verify:
   VM state has not changed after snapshot taking operation completes
   Todo: Vmdk creation?
   Todo: VM config has not changed after snapshot taking operation completes

Revert to snapshot from these VM power states:
   poweredOn: Revert to SNPoweredOn
   suspended: Revert to SNPoweredOn
   poweredOff: Revert to SNPoweredOn
  Verify:
   VM state has changed to the state captured by snapshot.
   Todo: If VM is a real OS and support file operations, Power on VM to verify Snapshot ID file in VM.
   Todo: VM config has changed to the state captured by snapshot. Where to store VM config information of Snapshot?

Delete snapshot in these VM power states:
   poweredOn: Delete SNPoweredOn
   suspended: Delete SNPoweredOn
   poweredOff: Delete SNPoweredOn
  Verify:
   VM state has not changed after snapshot deleting operation completes
   Todo: Snapshot object is deleted.
   Todo: Vmdk file deletion? disk cosolidation?
   Todo: VM config has not changed after snapshot deleting operation completes

Delete snapshot with/without consolidation:
   True
   False
  Optional: Verify consolidation flag after deletion complete.

Remove all snapshots with/without consolidation:
  Verify:
   VM state has not changed after snapshot deleting operation completes
   Todo: All snapshots are deleted.
   Todo: Vmdk file deletion? disk cosolidation?
   Todo: VM config has not changed after snapshot deleting operation completes

Create a width n depth m snapshot tree

Todo: Limit line length to 79 max.
Todo: Revert to specific snapshot in width n depth m snapshot tree
Todo: Delete specific snapshot in width n depth m snapshot tree
Todo: Make add/remove test case easy
Todo: Create testcases for combination of disk modes, e.g. persistent, independent.
Todo: Create testcases for snapshot operation with various VM configurations, e.g. Added virtual Nics or other virtual devices
'''

from __future__ import print_function

import sys
from pyVmomi import Vim
#from pyVim.connect import Connect, Disconnect
from pyVim.connect import SmartConnect, Disconnect, GetSi
from pyVim.task import WaitForTask
from pyVim import folder
from pyVim import vm
from pyVim import invt
from pyVim import vmconfig
from pyVim import vimutil
from pyVim import arguments
from pyVim.helpers import Log,StopWatch
import time
import atexit

import datetime

class TestFlow:
   def __init__(self):
      self.TestNumber = 0
      self.TestClock = StopWatch()
      self.TestResults = []

   def SetPhase(self, testname, testresult):
      Log("Test Case %s %s completed: %s" % (self.TestNumber, testname, testresult))
      self.TestClock.finish("Test Case %s" % self.TestNumber)
      self.timeTaken = self.TestClock.timeTaken
      self.TestClock = StopWatch()
      self.TestResults.append((testname, testresult, self.timeTaken))

   def testCreateSingleSnapshot(self, namebase, desc, memory, quiesce, vm1, vmstate, listSnapshots):
      self.TestNumber = self.TestNumber + 1
      testtype = "Snapshot.Create"
      testdesc = "Create Snapshot"

      # Construct Snapshot Name with namebase, vm state, memory, quiesce status
      memStr = ""
      qscStr = ""
      if memory:
         memStr = "Mem"
      if quiesce:
         qscStr = "Qsc"
      snapshotname = "%s%s%s%s" % (namebase, vmstate, memStr, qscStr)

      testdesc = "%s %s while VM is %s" % (testdesc, snapshotname, vmstate)
      Log("Test Case # %s: %s" % (self.TestNumber, testdesc))
      teststatus = "PASS"

      listSnapshots.append(SnapshotObj(snapshotname, desc, memory, quiesce, vm1, vmstate))
      if not listSnapshots[-1].snapref:
         Log("Error populating snapshot object.")
         teststatus = "FAIL"
      testcasename = "%s_%s" % (testtype, snapshotname)
      self.SetPhase(testcasename, teststatus)

   def testCreateMultipleSnapshots(self, namebase, vm1, vmstate, width=1, depth=1, memory=True, quiesce=False):
      #
      # Create same type of snapshots of defined depth and width
      # Example of width 4, depth 3 snapshot tree:
      # SN-SN_01-SN_02
      # |-SN01-SN01_01-SN01_02
      # |-SN02-SN02_01-SN02_02
      # |-SN03-SN03_01-SN03_02
      #
      self.TestNumber = self.TestNumber + 1
      testtype = "Snapshot.CreateMultiple%sx%s" % (width, depth)
      testdesc = "Create Multiple Snapshot %s in width %s in depth while VM is %s" % (width, depth, vmstate)

      # The depth of snapshots are limited to 31. If try to create depth of 32 snapshot,
      # You will run into following error:
      # Create snapshot SNpoweredOnMemQsc_31 of SnapshotTestVM VM failed:
      # (vim.fault.TooManySnapshotLevels) {
      #  dynamicType = <unset>,
      #  dynamicProperty = (vmodl.DynamicProperty) [],
      #  msg = 'Snapshot hierarchy is too deep.',
      #  faultCause = <unset>,
      #  faultMessage = (vmodl.LocalizableMessage) []}
      #
      # The depth > 31, this test will be transfer into Snapshot depth limit test.
      #

      # Construct Snapshot Name with namebase, vm state, memory, quiesce status
      memStr = ""
      qscStr = ""
      if memory:
         memStr = "Mem"
      if quiesce:
         qscStr = "Qsc"
      basename = "%s%s%s%s" % (namebase, vmstate,
                                         memStr, qscStr)

      #
      # XXX - We are going to retrieve the real snapshot tree information to compare with our snapshot array.
      # Currently, we expect no existing snapshot when we perform the limit test.
      #
      if depth > 31:
         testtype = "Snapshot.CreateMultiple%sx%sTestDepthLimitOnBranch%s" % (width, depth, width)
         testdesc = "Create Multiple Snapshot %s in width %s in depth, Verify 32th depth snapshot cannot be created on branch %s" % (width, depth, width)
         Log("Test Case # %s: %s" % (self.TestNumber, testdesc))
         teststatus = "PASS"
         widthSnapshots = [None for w in range(width - 1)]
         branchSnapshots = [None for d in range(depth)]
         # Create root snapshot (level 1 for every branch)
         snapshotname = basename
         Log("Create root snapshot %s." % snapshotname)
         branchSnapshots[0] = SnapshotObj(snapshotname, snapshotname,
                            memory, quiesce, vm1, vmstate)
         if not branchSnapshots[0]:
            teststatus = "FAIL"
            Log("Cannot create root snapshot")
            return None

         # Create level 2 snapshot for width-1 branches, and rever to rootSnap
         for k in range(width - 1):
            snapshotname = "%s%s" % (basename, k)
            widthSnapshots[k] = SnapshotObj(snapshotname, snapshotname,
                            memory, quiesce, vm1, vmstate)
            if not widthSnapshots[k]:
               teststatus = "FAIL"
               Log("Cannot create branch %s level 2 snapshot" % k + 1)
               return None
            if not revertSnapshot(vm1, branchSnapshots[0]):
               teststatus = "FAIL"
               Log("Cannot revert to root snapshot to create branch %s" % k + 2)
               return None

         try:
            # Create level 2 to level depth snapshot on target branch
            lastDepth = 1
            for j in range(depth - 1):
               snapshotname = "%s%s_%s" % (basename, width, j + 2)
               Log("Create snapshot %s at level %s." % (snapshotname, j + 2))
               branchSnapshots[j+1] = SnapshotObj(snapshotname, snapshotname,
                               memory, quiesce, vm1, vmstate)
               if not branchSnapshots[j+1]:
                  Log("Cannot create branch %s level %s snapshot" % (width, j + 2))
                  teststatus = "FAIL"
                  break
               lastDepth = j + 2

         except Exception as e:
            print(e)
            if "Snapshot hierarchy is too deep" in e.msg:
               if lastDepth == 31:
                  Log("Create snapshot %s of depth %s has failed as expected: %s" % (snapshotname, lastDepth + 1, e))
                  teststatus = "PASS"
               elif lastDepth > 31:
                  Log("Create snapshot %s of depth %s has failed, however exceed the depth limit 32: %s" % (snapshotname, lastDepth + 1, e))
                  teststatus = "FAIL"
               else:
                  Log("Create snapshot %s of depth %s has failed before the depth limit is reached: %s" % (snapshotname, lastDepth + 1, e))
                  teststatus = "FAIL"
            else:
               Log("Error: Create snapshot %s of depth %s has failed not for expected reason: %s" % (snapshotname, lastDepth + 1, e))
               teststatus = "FAIL"

         testcasename = "%s_%s" % (testtype, namebase)
         self.SetPhase(testcasename, teststatus)

         if teststatus == "PASS":
            return branchSnapshots + widthSnapshots

         return None

      # Test matrix of width x depth snapshot when depth <= 31
      treeSnapshots = [ [None for d in range(depth) ] for w in range(width)]
      Log("Test Case # %s: %s" % (self.TestNumber, testdesc))
      teststatus = "PASS"

      for j in range(width):

         if j:
            # Revert the snapshot tree root to start a new branch
            if not revertSnapshot(vm1, treeSnapshots[0][0]):
               teststatus = "FAIL"
               break

         for i in range(depth):
            if i > 0 and i <10:
               print("here is %s" % i)
               depStr = "_0%s" % i
            elif i == 0:
               depStr = ""
            else:
               depStr = "_%s" % i

            if j > 0 and j <10:
               print("here is %s" % j)
               widStr = "0%s" % j
            elif j == 0:
               widStr = ""
            else:
               widStr = "%s" % j

            snapshotname = "%s%s%s%s%s%s" % (namebase, vmstate,
                                         memStr, qscStr, widStr, depStr)
            try:
               treeSnapshots[j][i] = SnapshotObj(snapshotname, snapshotname,
                                               memory, quiesce, vm1, vmstate)
            except Exception as e:
               Log("Error: Create snapshot %s of depth %s has failed not for expected reason: %s" % (snapshotname, i + 1, e))
               teststatus = "FAIL"
               break
            if not treeSnapshots[j][i]:
               Log("Error populating snapshot object.")
               teststatus = "FAIL"

      testcasename = "%s_%s" % (testtype, namebase)
      self.SetPhase(testcasename, teststatus)

      k = 0
      for j in range(width):
         for i in range(depth):
            if treeSnapshots[j][i]:
               k = k + 1
               print("Created Snapshot branch %s depth %s: %s" % (j, i, treeSnapshots[j][i].name))
      print("Total of %s snapshots created in test" % k)

      if teststatus == "PASS":
         return treeSnapshots

      return None

   def testRevertToSnapshot(self, snap, desc, vm1, vmstate):
      self.TestNumber = self.TestNumber + 1
      testtype = "Snapshot.Revert"
      testdesc = "Revert to Snapshot"
      testdesc = "%s %s while VM is %s" % (testdesc, snap.name, vmstate)
      Log("Test Case # %s: %s" % (self.TestNumber, testdesc))
      teststatus = "FAIL"
      if revertSnapshot(vm1, snap):
         teststatus = "PASS"
      testcasename = "%s_VM%s_%s" % (testtype, vmstate, snap.name)
      self.SetPhase(testcasename, teststatus)

   def testDeleteSnapshot(self, snap, desc, vm1, vmstate, consolidate):
      self.TestNumber = self.TestNumber + 1
      testdesc = "Delete Snapshot"
      if consolidate:
         testtype = "Snapshot.DeleteConsolidate"
         testdesc = "%s %s with consolidate while VM is %s" % (testdesc, snap.name, vmstate)
      else:
         testtype = "Snapshot.Delete"
         testdesc = "%s %s without consolidate while VM is %s" % (testdesc, snap.name, vmstate)
      Log("Test Case # %s: %s" % (self.TestNumber, testdesc))
      teststatus = "FAIL"
      if deleteSnapshot(vm1, snap, consolidate):
         teststatus = "PASS"
      testcasename = "%s_VM%s_%s" % (testtype, vmstate, snap.name)
      self.SetPhase(testcasename, teststatus)

   def testRemoveAllSnapshots(self, vm1, vmstate, consolidate):
      self.TestNumber = self.TestNumber + 1
      testdesc = "Remove All Snapshot"
      if consolidate:
         testtype = "Snapshot.RemoveAll"
         testdesc = "%s with consolidate while VM is %s" % (testdesc, vmstate)
      else:
         testtype = "Snapshot.RemoveAllConsolidate"
         testdesc = "%s without consolidate while VM is %s" % (testdesc, vmstate)
      Log("Test Case # %s: %s" % (self.TestNumber, testdesc))
      teststatus = "FAIL"
      if removeAllSnapshots(vm1, consolidate):
         teststatus = "PASS"
      testcasename = "%s_VM%s" % (testtype, vmstate)
      self.SetPhase(testcasename, teststatus)

def getPowerState(vm3):
   """
   Get a string representing the power state.

   @return - "poweredOn" if VM is on, "poweredOff" if VM is off, and others.
   """
   #Log(vm3.runtime.powerState)
   return vm3.runtime.powerState


def changeVMState(vm3, vmState):
   vmCurState = getPowerState(vm3)
   possibleVMStates = ["poweredOn", "suspended", "poweredOff"]
   if vmState not in possibleVMStates:
      Log("The target %s state is unexpected." % vmState)
      return False
   if vmCurState not in possibleVMStates:
      Log("The %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
      return False

   # PoweredOn state has ongoing changes
   # suspended and poweredOff while require VM changes to create new snapshot.
   #[ 2013-03-21 02:34:21.987747 ]Create snapshot SNsuspendedMemQscQsc of SnapshotTestVM VM failed: (vim.fault.SnapshotNoChange) {
   #dynamicType = <unset>,
   #dynamicProperty = (vmodl.DynamicProperty) [],
   #msg = 'Snapshot not taken since the state of the virtual machine has not changed since the last snapshot operation.',
   #faultCause = <unset>,
   #faultMessage = (vmodl.LocalizableMessage) [
   #   (vmodl.LocalizableMessage) {
   #      dynamicType = <unset>,
   #      dynamicProperty = (vmodl.DynamicProperty) [],
   #      key = 'msg.snapshot.vigor.take.error',
   #      arg = (vmodl.KeyAnyValue) [
   #         (vmodl.KeyAnyValue) {
   #            dynamicType = <unset>,
   #            dynamicProperty = (vmodl.DynamicProperty) [],
   #            key = '1',
   #            value = 'msg.snapshot.error-NOCHANGE'
   #         }
   #      ],
   #      message = 'An error occurred while taking a snapshot: msg.snapshot.error-NOCHANGE.'
   #   }
   #]
   #}

   Log("The %s VM is in %s state, try change it to %s state:" % (vm3.config.name, vmCurState, vmState))
   if vmCurState == "poweredOn":
      if vmState == "poweredOff":
         if powerOff(vm3):
            Log("The %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
      elif vmState == "suspended":
         if suspend(vm3):
            Log("The %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
      elif vmState == "poweredOn":
         Log("The %s VM remains in %s state:" % (vm3.config.name, vmState))
         return True
   elif vmCurState == 'poweredOff':
      if vmState == "poweredOn":
         if powerOn(vm3):
            Log("The %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
      elif vmState == "suspended":
         if powerOn(vm3):
            Log("Power on %s VM." % (vm3.config.name))
         if suspend(vm3):
            Log("Then suspend %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
      elif vmState == "poweredOff":
         if powerOn(vm3):
            Log("Power on %s VM." % (vm3.config.name))
         if powerOff(vm3):
            Log("Then power off %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
   elif vmCurState == 'suspended':
      if vmState == "poweredOn":
         if powerOn(vm3):
            Log("The %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
      elif vmState == "poweredOff":
         if powerOff(vm3):
            Log("The %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True
      elif vmState == "suspended":
         if powerOn(vm3):
            Log("Power on %s VM to change state." % (vm3.config.name))
         if suspend(vm3):
            Log("Then suspend %s VM successfully changed to %s state:" % (vm3.config.name, vmState))
            return True

   Log("Unhandled VM state change condition, Error.")
   return False

def powerOff(vm3):
   """
   Power off the VM (hard power off, no guest shutdown)

   @return - False if VM not powered off, True if powered off.
   """

   vmState = getPowerState(vm3)

   if not vmState in ['poweredOn', 'suspended']:
      Log("The %s VM is not in state which allow power off operation: %s" % (vm3.config.name, vmState))
      return False

   try:
      Log("Powering off %s VM from %s state" % (vm3.config.name, vmState))
      task = vm3.PowerOff()
      WaitForTask(task)
      if getPowerState(vm3) == 'poweredOff':
         Log("The %s VM is powered off" % vm3.config.name)
         return True

   except Exception as e:
      Log("Power off %s VM failed: %s" % (vm3.config.name, e))

   Log("Failed to power off %s VM." % vm3.config.name)
   return False

def powerOn(vm3):
   """
   Power on the VM

   @return - False if VM not powered on, True if powered on.

   """

   vmState = getPowerState(vm3)

   if not vmState in ('poweredOff','suspended'):
      Log("The %s VM is not in a powered off/suspend state: %s" % (vm3.config.name, vmState))
      return False

   try:
      Log("Powering on %s VM from %s state" % (vm3.config.name, vmState))
      task = vm3.PowerOn()
      WaitForTask(task)

      if getPowerState(vm3) == 'poweredOn':
         Log("The %s VM is powered on" % vm3.config.name)
         return True

   except Exception as e:
      Log("Power on %s VM failed: %s" % (vm3.config.name, e))

   Log("Failed to power on %s VM" % vm3.config.name)
   return False

def suspend(vm3):
   """
   Suspend VM

   @param - vm [in] VirtualMachine object

   @return - False if VM not suspend, True if suspend.
   """

   vmState = getPowerState(vm3)

   if not vmState in ('poweredOn'):
      Log("The %s VM is not in a powered on state to suspend: %s" % (vm3.config.name, vmState))
      return False

   Log("Suspend %s VM from %s state" % (vm3.config.name, vmState))
   try:
      task = vm3.Suspend()
      WaitForTask(task)

      if getPowerState(vm3) == 'suspended':
         Log("The %s VM is suspended" % vm3.config.name)
         return True

   except Exception as e:
      Log("Suspend %s VM from %s state failed: %s" % (vm3.config.name, vmState, e))

   Log("Failed to suspend %s VM" % vm3.config.name)
   return False

def takeSnapShot(vm3, vmState, ssName="SNTest", ssDesc="SNTest", withMem=True, withQuiesce=False, verify=True):
   ssTypeStr = ""
   if withMem:
      ssTypeStr += "withMemory"
   if withQuiesce:
      ssTypeStr += " withQuiesce"
   if not changeVMState(vm3, vmState):
      Log("VM not in expected state before taking snapshot.")
      raise Exception("Error: Change VM State failed, VM not in expected state before taking snapshot.")

   Log("Create %s snapshot %s<%s> on %s VM when VM is %s" % (ssTypeStr, ssName, ssDesc, vm3.config.name, vmState))

   try:
      task = vm3.CreateSnapshot(ssName, ssDesc, withMem, withQuiesce)
      WaitForTask(task)
      vmStateAfter = getPowerState(vm3)
      # Getting snapshot obj from currentSnapshot is not recommended
      # It may not work all the time,
      # especially with quiesced snapshots.
      # Use task.info.result please.
      # snap = vm3.snapshot.currentSnapshot
      snap = task.info.result
      Log("Snapshot %s creation completed." % repr(snap))

      if vmState != vmStateAfter:
         Log("Error the VM state %s changed to %s after creating snapshot is not expected." % (vmState, vmStateAfter))
         raise Exception("Error: the VM state %s changed to %s after creating snapshot is not expected." % (vmState, vmStateAfter))

      if getPowerState(vm3) == "poweredOn":

         if verify:
            Log("Let's do some verification in Guest. Currently nothing.")

      return snap

   except Exception as e:
      Log("Create snapshot %s of %s VM failed: %s" % (ssName, vm3.config.name, e))
      #if "Snapshot hierarchy is too deep" in e.msg:
      raise

   Log("Failed to create snapshot %s of %s VM" % (ssName, vm3.config.name))
   raise Exception("Error: Failed to create snapshot %s of %s VM" % (ssName, vm3.config.name))

class SnapshotObj:
    def __init__(self, name, desc, memory, quiesce, vm, vmstate):
        self.name = name
        self.desc = desc
        self.memory = memory
        self.quiesce = quiesce
        self.vmstate = vmstate
        if vmstate == "poweredOn" and not memory:
           self.snstate = "poweredOff"
        else:
           self.snstate = vmstate
        self.vm = vm
        self.snapref = takeSnapShot(vm, vmstate, ssName=name, ssDesc=desc, withMem=memory, withQuiesce=quiesce)
        self.fileID = name

def createListOfSnapshots(vm1, cleanup=True):
   i = 0
   listSnapshots = []
   if cleanup:
      try:
         Log("Cleanup existing snapshot before creating new list of snapshots.")
         Log("Remove all snapshots and consolide")
         vm.RemoveAllSnapshots(vm1)
      except Exception as e:
         Log("Error when remove all snapshot: %s" % e)
         raise

   testnamebase = "SN"
   vmstate = "poweredOn"
   testname = ""
   for memory in [True, False]:
      for quiesce in [True, False]:
         snapshotTypeStr = ""
         if memory:
            snapshotTypeStr += "Mem"
         if quiesce:
            snapshotTypeStr += "Qsc"
         testname = "%s%s%s" % (testnamebase, vmstate, snapshotTypeStr)
         i = i + 1
         Log("Create Snapshot %s %s" % (i, testname))
         listSnapshots.append(SnapshotObj(testname, testname, memory, quiesce, vm1, vmstate))
         if not listSnapshots[-1].snapref:
            Log("Error when create Snapshot %s %s" % (i, testname))
            return False

   for vmstate in ["suspended", "poweredOff"]:
      memory = False
      quiesce = False
      testname = "%s%s" % (testnamebase, vmstate)
      i = i + 1
      Log("Create Snapshot %s %s" % (i, testname))
      listSnapshots.append(SnapshotObj(testname, testname, memory, quiesce, vm1, vmstate))
      if not listSnapshots[-1].snapref:
         Log("Error when create Snapshot %s %s" % (i, testname))
         return False

   return listSnapshots

def revertSnapshot(vm1, snap):
   try:
      task = snap.snapref.Revert()
      WaitForTask(task)
   except Exception as e:
      Log("Error when reverting to snapshot %s: %s" % (snap.name, e))
      return False
   vmState = getPowerState(vm1)
   if vmState != snap.snstate:
      Log("Error the VM state %s after reverting to snapshot %s is not expected." % (snap.snstate, snap.name))
      return False
   return True

def deleteSnapshot(vm1, snap, consolidate=True):
   vmstate = getPowerState(vm1)
   if consolidate:
      try:
         vimutil.InvokeAndTrack(snap.snapref.Remove, False)
         # This consolidationNeeded check is not necessarily correct.
         # We don't guarantee that consolidate will succeed when a snapshot is removed.
         # Let's investigate before making it a pass/fail critira.
         # (i.e. Snapshot.Remove might succeed, but the VM still needs more consolidating.
         #  VM may be left in consolidationNeeded True state when other snapshot is removed.
         #  Currently consolidation will timeout if it cannot complete within 12sec.)
         # Giving a warning here instead of raise exception.
         if vm1.runtime.consolidationNeeded:
            Log("Warning: Consolidation is likely failed while removing snapshot %s. VM still has consolidationNeeded flag set." % snap.name)
      except Exception as e:
         Log("Error when deleting snapshot %s: %s" % (snap.name, e))
         return False

   else:
      try:
         vimutil.InvokeAndTrack(snap.snapref.Remove, False, False)
         # As commented above. We need to investigate the behavior of consolidationNeeded flag.
         # Giving a warning here instead of raise exception.
         if not vm1.runtime.consolidationNeeded:
            Log("Warning: Removing snapshot %s without consolidation. VM consolidationNeeded flag expected to be True but is cleared." % snap.name)
         Log("Consolidate snapshot disk is called on VM %s" % vm1.config.name)
         vimutil.InvokeAndTrack(vm1.ConsolidateDisks)
         # As commented above. We need to investigate the behavior of consolidationNeeded flag.
         # Giving a warning here instead of raise exception.
         if vm1.runtime.consolidationNeeded:
            Log("Warning: Consolidation flag not cleared after consolidate disks. It is likely failed or timed out.")
      except Exception as e:
         Log("Error when deleting to snapshot %s: %s" % (snap.name, e))
         return False

   vmStateAfter = getPowerState(vm1)
   if vmStateAfter != vmstate:
      Log("Error the VM state %s changed to %s after deleting snapshot %s is not expected." % (vmstate, vmStateAfter, snap.name))
      return False

   return True

def removeAllSnapshots(vm1, consolidate=True):
   # Let's enforce our test is testing removal of existing snapshot.
   if not vm1.snapshot.currentSnapshot:
      Log("No snapshot exist before remove all operation. \
           Please check your test design to make sure snapshot exist before removal.")
      return False
   vmstate = getPowerState(vm1)
   if consolidate:
      try:
         Log("Remove all snapshots and consolide")
         vm.RemoveAllSnapshots(vm1)
         # As commented above. We need to investigate the behavior of consolidationNeeded flag.
         # Giving a warning here instead of raise exception.
         if vm1.runtime.consolidationNeeded:
            Log("Warning: consolidationNeeded flag not cleared after removeall snapshots with consolidation. Expect VM no longer need consolidation.")
      except Exception as e:
         Log("Error when remove all snapshot: %s" % e)
         return False
   else:
      try:
         Log("Remove all snapshots without consolide")
         vimutil.InvokeAndTrack(vm1.RemoveAllSnapshots, False)
         # As commented above. We need to investigate the behavior of consolidationNeeded flag.
         # Giving a warning here instead of raise exception.
         if vm1.runtime.consolidationNeeded:
            Log("Warning: consolidationNeeded flag is False after removeall snapshots without consolidation. Expect VM still need consolidation.")
      except Exception as e:
         Log("Error when deleting to snapshot %s: %s" % (snap.name, e))
         return False
   vmStateAfter = getPowerState(vm1)
   if vmStateAfter != vmstate:
      Log("Error the VM state changed from %s to %s after deleting all snapshot is not expected." % (vmstate, vmStateAfter))
      return False

   return True

def main():
   supportedArgs = [ (["h:", "host="], "localhost", "Host name", "host"),
                     (["u:", "user="], "root", "User name", "user"),
                     (["p:", "pwd="], "", "Password", "pwd"),
                     (["v:", "vmname="], None, "Name of the virtual machine", "vmname"),
                     (["d:", "datastore="], None, "datastore", "datastore"),
                     (["v:", "volume="], None, "VMFS volume", "volume"),
                     (["r:", "regression="], "all", "Regression test group sanity|precheckin|all", "regression"),
                     (["i:", "iter="], "1", "Num of iterations", "iter")]
   supportedToggles = [(["usage", "help"], False, "Show usage information", "usage")]

   args = arguments.Arguments(sys.argv, supportedArgs, supportedToggles)
   if args.GetKeyValue("usage") == True:
      args.Usage()
      sys.exit(0)

   # Connect
   si = SmartConnect(host=args.GetKeyValue("host"),
                     user=args.GetKeyValue("user"),
                     pwd=args.GetKeyValue("pwd"),
                     port=443)
   atexit.register(Disconnect, si)

   about = si.content.about
   Log("The connection product line information %s" % about.productLineId)

   # Process command line
   vmname = args.GetKeyValue("vmname")
   numiter = int(args.GetKeyValue("iter"))
   datastore = args.GetKeyValue("datastore")
   volume = args.GetKeyValue("volume")
   regression = args.GetKeyValue("regression")

   #
   # Try to group the tests by priority
   #  sanity - high priority baseline tests, complete in 10 minutes (dummy VM)
   #  precheckin - important tests to be run as precheckin, complete in 30 minutes (dummy VM)
   #  all - all tests, complete in 2hr (dummy VM)
   #
   regType = ["sanity", "precheckin", "all"]
   if regression in regType:
      Log("Run %s regression tests." % regression)
   else:
      Log("Error: Regression type %s is not defined. Exit." % regression)
      args.Usage()
      sys.exit(0)

   start = datetime.datetime.now()

   #
   # Setup
   #
   testVMName = "SnapshotTestVM"
   if vmname:
      # Cleanup from previous runs.
      vm1 = folder.Find(vmname)
      if vm1 == None:
         print("VM not found.")
         Log("Let's create a dymmy VM to run snapshot tests.")
         # Create a simple vm
         vm1 = vm.CreateQuickDummy(testVMName, 1, datastoreName=datastore)
         Log("Dummy VM %s is created." % testVMName)
      else:
         Log("Found VM %s to perform test, verify it is in power off state." % vm1.config.name)
         if (vm1.runtime.powerState != Vim.VirtualMachine.PowerState.poweredOff):
            try:
               WaitForTask(vm1.PowerOff())
            except: pass
   else:
      vm1 = folder.Find(testVMName)
      if vm1 == None:
         # Create a simple vm
         vm1 = vm.CreateQuickDummy(testVMName, 1, datastoreName=datastore)
         Log("Dummy VM %s is created." % testVMName)
      else:
         try:
            WaitForTask(vm1.PowerOff())
         except: pass
         Log("Please cleanup unit test vms from previous runs. Test results may be affected.")
   Log("Remove all snapshots and consolide before test starts.")
   vm.RemoveAllSnapshots(vm1)

   tc = TestFlow()
   for j in xrange(numiter) :

      listSnapshots = []

      # Sanity tests
      if regression in ["sanity", "precheckin", "all"]:
         #
         # Test Cases of Creating Snapshots with different options
         #
         vmstate = "poweredOn"
         for memory in [True, False]:
            for quiesce in [True, False]:
               memStr = ""
               qscStr = ""
               if memory:
                  memStr = "Mem"
               if quiesce:
                  qscStr = "Qsc"
               testdesc = "SN%s%s%s" % (vmstate, memStr, qscStr)
               tc.testCreateSingleSnapshot("SN", testdesc, memory, quiesce, vm1, vmstate, listSnapshots)

         memory = False
         quiesce = False
         for vmstate in ["suspended", "poweredOff"]:
            testdesc = "SN%s" % vmstate
            tc.testCreateSingleSnapshot("SN", testdesc, memory, quiesce, vm1, vmstate, listSnapshots)

         #
         # Test Cases of Reverting Snapshots to different VM states
         #
         for vmstate in ["poweredOn", "suspended", "poweredOff"]:
            changeVMState(vm1, vmstate)

            for snap in listSnapshots:
               tc.testRevertToSnapshot(snap, snap.name, vm1, vmstate)

         #
         # Test Cases of Deleting Snapshots while VM is in different state
         # With consolidation
         #

         changeVMState(vm1, "poweredOn")

         for snap in listSnapshots:
            # Delete snapshot with consolidate
            tc.testDeleteSnapshot(snap, snap.name, vm1, vmstate, True)

      # precheck in tests
      if regression in ["precheckin", "all"]:

         #
         # Test Cases of Deleting Snapshots while VM is in different state
         # Without consolidation
         #

         for vmstate in ["poweredOn", "suspended", "poweredOff"]:
            newlistSnapshots = createListOfSnapshots(vm1)
            changeVMState(vm1, vmstate)
            for snap in newlistSnapshots:
               # Delete snapshot without consolidate
               tc.testDeleteSnapshot(snap, snap.name, vm1, vmstate, False)

      #
      # Test Cases of Removing All Snapshots while VM is in different state
      # Without consolidation
      #

      # Sanity tests
      if regression in ["sanity", "precheckin", "all"]:
         listSnapshots = createListOfSnapshots(vm1)
         changeVMState(vm1, "poweredOn")
         # Remove all snapshot with consolidate
         tc.testRemoveAllSnapshots(vm1, "poweredOn", False)

      # precheck in tests
      if regression in ["precheckin", "all"]:
         for vmstate in ["suspended", "poweredOff"]:
            listSnapshots = createListOfSnapshots(vm1)
            changeVMState(vm1, vmstate)
            # Remove all snapshot with consolidate
            tc.testRemoveAllSnapshots(vm1, vmstate, False)

      # precheck in tests
      if regression in ["precheckin", "all"]:

         #
         # Test Cases of Creating multiple Snapshots to the depth limit
         # Verify 31 snapshots can be created in one branch, 32th cannot be created.
         # Please clean up all existing snapshot before run this test.
         #
         tc.testCreateMultipleSnapshots("SN", vm1, "poweredOn", 3, 32, True, True)

         #
         # Test Cases of Removing All Snapshots while VM is in different state
         # With consolidation
         #

         # Remove all snapshot with consolidate in poweron state
         tc.testRemoveAllSnapshots(vm1, "poweredOn", True)

      # low priority tests
      if regression in ["all"]:

         #
         # Test Cases of Creating multiple Snapshots of width 10
         #
         tc.testCreateMultipleSnapshots("SN", vm1, "poweredOn", 10, 1, True, True)

         changeVMState(vm1, "suspended")
         # Remove all snapshot without consolidate
         tc.testRemoveAllSnapshots(vm1, "suspended", True)

         #
         # Test Cases of Creating multiple Snapshots 2x2
         #
         tc.testCreateMultipleSnapshots("SN", vm1, "poweredOn", 2, 2, False, False)
         tc.testCreateMultipleSnapshots("SN", vm1, "suspended", 2, 2, False, False)
         tc.testCreateMultipleSnapshots("SN", vm1, "poweredOff", 2, 2, False, False)

         changeVMState(vm1, "poweredOff")
         # Remove all snapshot without consolidate
         tc.testRemoveAllSnapshots(vm1, "poweredOff", True)

      Log("Success: test iter %s" % j)

   end = datetime.datetime.now()
   k = 0
   for test in tc.TestResults:
      k = k + 1
      print("Test %s %s: %s %s" % (k, test[0], test[1], test[2]))
   duration = end - start
   print("Total time taken: %s" % duration)

# Start program
if __name__ == "__main__":
   main()
